#
# CVSWrapper
#
# A python wrapper class for the cvs command line tool
#
# Copyright (c) 2001 by beehive elektronische medien GmbH
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public License
# (MPL)
#

"""
$Id: CVSWrapper.py,v 1.12 2003/05/02 13:58:02 philipp Exp $
"""

import sys
import os
import string
import re

from ShellCommandProcessor import ShellCommandProcessor
from scramble import scramble
from interfaces.RevisionController import IRevisionController

# unused yet maybe useful later
_cvs_states = {
    'Up-to-date':       'OK',
    'Locally Modified': 'LM',
    'Locally Added':    'LA',
    'Locally Removed':  'LR',
    'Needs Checkout':   'NC',
    'Needs Patch':      'NP',
    'Unknown':          'UN'
    }

_update_flags = {
    'U': 'Updated',
    'P': 'Patched',
    'A': 'Locally Added',
    'R': 'Locally Removed',
    'M': 'Locally Modified',
    'C': 'File had conflicts on merge',
    '?': 'Unknown'
    }

_import_flags = {
    'U': 'File already existed',
    'N': 'New file added',
    'C': 'Locally Modified',
    'I': 'Ignored',
    'L': 'Symbolic link'
    }

_tag_flags = {
    'T': 'Tagged',
    'D': 'Tag deleted'
    }

_cvs_commands = ("init", "status", "checkout", "commit", "add", "remove",
                 "update", "import", "tag", "log", "diff")

BAD_TAGS = ("BASE", "HEAD")

#
# see http://www.uni-ulm.de/~s_smasch/various/argmax/
#
ARG_MAX = 64*1024

class CVSError(Exception):
    pass

class CVSWrapper:
    """
    Wrapper class for the cvs command. Makes nearly all cvs functionality
    available as instance methods.
    """

    __implements__ = IRevisionController

    def __init__(self, **kw):
        # required keyword arguments
        try:
            repository_path = kw['repository_path']
            module          = kw['module']
            work_dir        = kw['work_dir']
        except:
            raise ValueError, "Insufficient parameters."
        if not (repository_path and module and work_dir):
            raise ValueError, "Insufficient parameter values."

        self.repository_path = repository_path
        self.module          = module
        self.work_dir        = work_dir

        # optional but helpful...
        self.cvs_bin         = kw.get('cvs_bin', 'cvs')
        self.access_method   = kw.get('access_method', 'local')
        self.server_addr     = kw.get('server_addr', 'localhost')
        self.user_name       = kw.get('user_name', '')
        self.passwd          = kw.get('passwd', '')
        self.cvs_rsh         = kw.get('cvs_rsh', 'rsh')
        self.comp_level      = kw.get('comp_level', '3')
        self.cvs_passfile    = kw.get('cvs_passfile', '')
        self.debug           = kw.get('debug', 0)

        # assemble repository string
        repository = ":%s:" % self.access_method
        if self.access_method != "local" and self.server_addr:
            if self.user_name:
                repository += "%s@" % self.user_name
            repository += "%s:" % self.server_addr
        repository += self.repository_path
        self.repository = repository
        
        # generate environment dict for later use
        environment = {}
        if self.access_method == "pserver":
            environment['CVS_PASSFILE'] = self.cvs_passfile
        if self.access_method in ("ext", "server") and self.cvs_rsh:
            environment['CVS_RSH'] = self.cvs_rsh
        self.environment = environment

        if self.access_method == "pserver":
            self._login(passwd)

    def __del__(self):
        if hasattr(self, "access_method"):
            if self.access_method == "pserver":
                self._logout()

    def _login(self, passwd):
        """
        Perform a CVS login
        
        This does not really login but writes the correspondent line to a
        '.cvspass'-like file
        """
        cvspassline = "%s %s\n" % (self.repository, scramble(passwd))
        fp = open(self.cvs_passfile, "a")
        fp.write(cvspassline)
        fp.close()

    def _logout(self):
        """
        perform a CVS logout - behaves just like login
        """
        lines = open(self.cvs_passfile, "r").readlines()
        newlines = filter(lambda l, r=self.repository:
                          l.find(r) == -1, lines)
        if newlines != lines:
            open(self.cvs_passfile, "w").writelines(newlines)
    
    def _is_logged_in(self):
        """
        Check whether we are still logged into cvs
        """
        if open(self.cvs_passfile, "r").read().find(self.repository) != -1:
            return 1
        return 0

    def _gen_cmd(self, cvscmd, cvsopt, cmdopt, filename):
        """
        Generate a cvs command string
        """
        if cvscmd not in _cvs_commands:
            self._clean_up()
            raise CVSError, "%s is not a supported cvs command" % command

        cmd = ""
        # first add environment variables
        for pair in self.environment.items():
            cmd += '%s="%s" ' % pair

        # repository
        cvsopt += "-d %s " % self.repository

        if not self.debug:
            cvsopt += "-q " # use quiet option when not in debug mode

        if self.comp_level and self.comp_level != "0":
            cvsopt += "-z%s " % self.comp_level

        cmd += "%s %s %s %s %s" % (self.cvs_bin, cvsopt, cvscmd, cmdopt,
                                   filename)
        return cmd
        
    def _exec_cmd(self, cvscmd, cvsopt="", cmdopt="", filename="",
                  raise_error=1):
        """
        Execute a command and process the streams
        """
        # don't go over 90% of ARG_MAX to be on the safe side
        max_filename_len = int(0.9*ARG_MAX)
        leftover = ""
        if len(filename) > max_filename_len:
            next_space = filename.find(' ', max_filename_len)
            filename, leftover = filename[:next_space], filename[next_space:]

        command = self._gen_cmd(cvscmd, cvsopt, cmdopt, filename)
        shellproc = ShellCommandProcessor()
        result_val, result_out, result_err = shellproc.processStreams(command)
        if self.debug:
            # debug mode simply means write result_err to sys.stderr
            try:
                sys.stderr.write(result_err)
            except IOError:
                # ouch - stderr is not writable
                pass
            
        if result_val and raise_error:
            # this is not portable, but very efficient
            self._clean_up()
            raise CVSError, result_err
        
        if leftover:
            # we still have leftover arguments
            leftover_val, leftover_out, leftover_err = \
                          self._exec_cmd(cvscmd, cvsopt, cmdopt, leftover,
                                         raise_error)
            result_val = result_val or leftover_val
            result_out += leftover_out
            result_err += leftover_err
        return result_val, result_out, result_err

    def _prepare(self):
        # change into the module directory
        self.olddir = os.getcwd()
        os.chdir(os.path.join(self.work_dir, self.module))

    def _clean_up(self):
        try:
            olddir = self.olddir
        except AttributeError:
            return
        # change back to the old directory
        os.chdir(olddir)
        del self.olddir

    ##########################################################################
    # Implementation methods for interface
    # interfaces.RevisionController.IRevisionController

    # standard CVS actions

    def initRepository(self):
        """
        initialize CVS repository
        only available if access method "local" has been chosen
        """
        if self.access_method != "local":
            raise CVSError, "Can only initialize local repositories"
        if not os.path.isdir(self.repository_path):
            os.mkdir(self.repository_path)
        result_val, result_out, result_err = self._exec_cmd("init")

    def status(self, filename="", recursive=1, verbose=0):
        """
        invoke CVS status and parse the output
        """
        self._prepare()
        cmdopt = ""
        if not recursive:
            cmdopt = "-l"
        if verbose:
            cmdopt = "-v " + cmdopt
        result_val, result_out, result_err = \
                    self._exec_cmd("status", "", cmdopt, filename)
        self._clean_up()
        return self.parse_status(result_out)

    def checkout(self, filename="", tag="", recursive=1, prune_empty=0):
        """
        perform a CVS checkout
        """
        # change into working directory
        olddir = os.getcwd()
        os.chdir(self.work_dir)

        cmdopt = self.module
        if not recursive:
            cmdopt = "-l " + cmdopt
        if prune_empty:
            cmdopt = "-P " + cmdopt
        if tag:
            cmdopt = "-r %s %s" % (tag, cmdopt)
        result_val, result_out, result_err = \
                    self._exec_cmd("checkout", "", cmdopt, filename)
        os.chdir(olddir)

        return self.parse_listout(result_out)

    def commit(self, log_message, filename="", recursive=1):
        """
        perform a CVS checkin/commit
        """
        self._prepare()
        cmdopt = '-m "%s"' % log_message
        if not recursive:
            cmdopt = "-l " + cmdopt
        result_val, result_out, result_err = \
                    self._exec_cmd("commit", "", cmdopt, filename)
        self._clean_up()
        return self.parse_listout(result_out)

    checkin = commit

    def add(self, filename, binary=0):
        """
        add a file to the repository
        """
        if not filename:
            raise CVSError, "You did specify a filename to add"
        self._prepare()
        cmdopt = ""
        if binary:
            cmdopt = "-kb"
        result_val, result_out, result_err = \
                    self._exec_cmd("add", "", cmdopt, filename)
        self._clean_up()
        return result_out

    def remove(self, filename):
        """
        remove a file from the repository
        note: the actual file on the filesystem must NOT exist anymore
        """
        if not filename:
            raise CVSError, "You did not specify a filename to remove"
        self._prepare()
        result_val, result_out, result_err = \
                    self._exec_cmd("remove", filename=filename)
        self._clean_up()
        return result_out

    def update(self, filename="", tag="", recursive=1, prune_empty=0,
               new_dirs=0, reset_sticky=0):
        """
        perform a CVS update
        """
        self._prepare()
        cmdopt = ""
        if not recursive:
            cmdopt += "-l"
        if prune_empty:
            cmdopt += " -P"
        if new_dirs:
            cmdopt += " -d"
        if reset_sticky:
            cmdopt += " -A"
        if tag:
            cmdopt = "-r %s %s" % (tag, cmdopt)
        result_val, result_out, result_err = \
                    self._exec_cmd("update", "", cmdopt, filename)
        self._clean_up()

        updatedict = self.parse_listout(result_out)
        updatedict.update(self.parse_update_removed(result_err))
        return updatedict

    def importModule(self, vendor_tag, release_tag, log_message,
                     ignore_files=[], import_dir=None):
        """
        Do an initial import into the cvs repository
        """
        if import_dir is None:
            import_dir = os.path.join(self.work_dir, self.module)
        # change into working directory
        olddir = os.getcwd()
        os.chdir(import_dir)

        cmdopt = '-m "%s" ' % log_message
        for file in ignore_files:
           cmdopt += "-I %s " % file 
        cmdopt += "%s %s %s" % (self.module, vendor_tag, release_tag)
        
        result_val, result_out, result_err = \
                    self._exec_cmd("import", cmdopt=cmdopt)
        os.chdir(olddir)
        
        return self.parse_listout(result_out)

    def tag(self, tag, filename="", branch=0, recursive=1):
        """
        tag file(s)
        """
        # check for bad tag names
        if tag in BAD_TAGS:
            raise CVSError, "Bad tag name: %s" % tag
        
        # check for bad characters in tag
        allowed_chars = string.letters + string.digits + ".-_"
        for char in tag:
            if char not in allowed_chars:
                raise CVSError, "Bad character in tag: %s" % tag

        self._prepare()
        cmdopt = tag
        if branch:
            cmdopt = "-b " + cmdopt
        if not recursive:
            cmdopt = "-l " + cmdopt
        result_val, result_out, result_err = \
                    self._exec_cmd("tag", "", cmdopt, filename)
        self._clean_up()
        return result_out

    def log(self, filename="", recursive=1):
        """
        return log
        """
        self._prepare()
        cmdopt = ""
        if not recursive:
            cmdopt = "-l"
        result_val, result_out, result_err = \
                    self._exec_cmd("log", "", cmdopt, filename)
        self._clean_up()
        return self.parse_log(result_out)

    def diff(self, filename="", rev1="", rev2="", recursive=1):
        """
        Return a unified diff difference of a file between revision
        rev1 and revision rev2
        """
        if rev1:
           rev1 = "-r" + rev1
        if rev2:
           rev2 = "-r" + rev2

        self._prepare()
        #cmdopt = "%s %s -u" % (rev1, rev2) # unified patch format
        cmdopt = "%s %s" % (rev1, rev2)
        if not recursive:
            cmdopt = "-l " + cmdopt
        # cvs diff returns a positive return value if the revisions differ
        # thus, do not raise an exception on a positive return value.
        result_val, result_out, result_err = \
                    self._exec_cmd("diff", "", cmdopt, filename, raise_error=0)
        self._clean_up()
        
        # instead, raise an error if there is no diff but the return value is
        # still positive
        if not result_out and result_val:
            raise CVSError, result_err
        return result_out

    # end of standard CVS actions

    def merge(self, rev1, rev2, filename="", recursive=1):
        self._prepare()
        cmdopt = "-j %s -j %s" % (rev1, rev2)
        if not recursive:
            cmdopt = "-l " + cmdopt
        result_val, result_out, result_err = \
                    self._exec_cmd("update", "", cmdopt, filename)
        self._clean_up()
        return result_out

    def getLocallyModifiedFiles(self, filename="", recursive=1):
        # technically, we could use 'update' for this, but this would also
        # update files with newer revisions from the repository which is not
        # wanted. Therefore, we use a trick and abuse the output of the 'diff'
        # method.
        diff = self.diff(filename, recursive=recursive)
        pattern = re.compile("^Index: (.+)$", re.MULTILINE)
        return pattern.findall(diff)

    #
    ##########################################################################
    # parsers
    #

    def parse_status(self, output):
        """
        parse the output of the cvs status command
        also support the -v option for displaying tag information
        """
        returndict = {}
        patterns = [ r'File: (?P<filename>.+?)\s',
                     r'.+Status: (?P<status>.+?)\n',
                     r'.+Working revision:\s+(?P<wrev>.+?)\s',
                     r'.+Repository revision:\s+(?P<rrev>.+?)\s+(?P<rpath>.+?)\n',
                     r'.+Sticky Tag:\s+(?P<stag>.*?)\s*\(.*\)\n',
                     r'.+Sticky Date:\s+(?P<sdate>.+?)\n',
                     r'.+Sticky Options:\s+(?P<sopt>.+?)\n',
                     ]
        for i in range(len(patterns)):
            # DOTALL means the dot also matches a new line
            patterns[i] = re.compile(patterns[i], re.DOTALL)

        tag_pattern = r'\n\t(?P<tagname>\w+?)\s+\((?P<taglink>.+?)\)'
        tag_pattern = re.compile(tag_pattern)

        # length of the repository and module directory name
        try:
            rname_len = len(os.path.join(self.repository_path, self.module)) + 1
        except IndexError:
            rname_len = 0

        chunks = output.split(
            "===================================================================\n"
            )
        for chunk in chunks:
            chunkdict = {'tags': {}}
            for pattern in patterns:
                match = pattern.match(chunk)
                if match is not None:
                    chunkdict.update(match.groupdict())

            matches = tag_pattern.findall(chunk)
            tags = chunkdict['tags']
            for match in matches:
                tags[match[0]] = match[1]

            relfname = chunkdict.get('rpath', '')
            if relfname:
                relfname = relfname[rname_len:-2]
                returndict[relfname] = chunkdict
        return returndict
    
    #XXX redo this using regular expressions?!?
    def parse_log(self, log_output):
        """
        parse the output of the cvs log command
        """        
        returndict = {}
        chunks = log_output.split(
	    "=============================================================================\n"
	   )
        for fileinfo in chunks:
            d = { 'revisions': [] }
            infochunks = fileinfo.split("----------------------------\n")

            lines = infochunks[0].split('\n')
            if len(lines) < 11:
                continue

            lines = map(lambda line: line.split(), lines)
            relfname = lines[2][2]

            # parse main information
            try: d['head'] = lines[3][1]
            except IndexError: d['head'] = None
            try: d['branch'] = lines[4][1]
            except IndexError: d['branch'] = None
            try: d['locks'] = lines[5][1]
            except IndexError: d['locks'] = None
            try: d['access'] = lines[6][2]
            except IndexError: d['access'] = None

            # parse tags
            tags = {}
            for tagline in lines[8:-4]:
                tags[tagline[0][:-1]] = tagline[1]
            d['tags'] = tags

            # parse revisions
            for revinfo in infochunks[1:]:
                origlines = revinfo.split('\n')
                lines = map(lambda line: line.split(), origlines)
                rev = {
                    'revision': lines[0][1],
                    'date':     lines[1][1] + " " + lines[1][2][:-1],
                    'author':   lines[1][4][:-1],
                    'state':    lines[1][6][:-1],
                }
                if lines[2][0] == "branches:":
                    rev['log'] = "\n".join(origlines[3:-1])
                    rev['branches'] = origlines[2][11:-1]
                else:
                    rev['log'] = "\n".join(origlines[2:-1])
                    rev['branches'] = None

                # is there a +/- line info?
                if len(lines[1]) > 7:
                    rev['lines'] = lines[1][8] + " " + lines[1][9]
                else:
                    rev['lines'] = None

                d['revisions'].append(rev)
            returndict[relfname] = d
        return returndict

    def parse_listout(self, output, flagmapping=_update_flags):
        """
        parse the output of cvs update/checkout command
        """
        expr = "^([UPARMCINT?]) (.*)$"
        pattern = re.compile(expr)
        lines = output.split('\n')
        returndict = {}
        
        for line in lines:
            match = pattern.match(line)
            if match is not None:
                flag, filename = match.groups()
                flag = flagmapping.get(flag, flag)
                returndict[filename] = flag

        return returndict

    def parse_update_removed(self, output):
        """
        parse output of the 'update' command (usually the stderr
        output) and return all removed files
        """
        expr = "cvs update: (.+) is no longer in the repository"
        pattern = re.compile(expr)
        lines = output.split('\n')
        returndict = {}
        
        for line in lines:
            match = pattern.match(line)
            if match is not None:
                filename = match.groups()[0]
                returndict[filename] = "Removed from repository"

        return returndict
